
# single-hypergraph 
_single_datasets_= ['cora', 'citeseer', 'pubmed',
                    'coauthor_cora', 'coauthor_dblp',
                    '20newsW100', 'ModelNet40','NTU2012', 
                    'Mushroom','zoo',
                    'yelp','walmart-trips-100','house-committees-100','walmart-trips','house-committees',
                    'actor','amazon','pokec','twitch',
                    'bail','credit','german',
                    'amazon_review','magpm','magpm_mini',
                    'trivago','ogbn_mag']

# multi-hypergraph
_multi_datasets_=["RHG_3", "RHG_10", "RHG_table", "RHG_pyramid",
                  "IMDB_dir_form", "IMDB_dir_genre","IMDB_wri_form", "IMDB_wri_genre",
                  "stream_player","twitter_friend"]

# fair-hypergraph
_fair_datasets_ = ['credit','german','bail'] 